package com.isu.gaswellwatch.modbus.data.listener;

import cn.hutool.core.map.MapUtil;
import cn.hutool.json.JSONUtil;
import com.isu.gaswellwatch.modbus.data.ModbusMessage;
import com.isu.gaswellwatch.modbus.data.PersistenceHandler;
import com.isu.gaswellwatch.modbus.data.decode.DecodeHandler;
import com.isu.gaswellwatch.modbus.data.decode.impl.ComposeDecodeHandler;
import com.isu.gaswellwatch.modbus.data.decode.impl.HighLowBinDecodeHandler;
import com.isu.gaswellwatch.modbus.data.decode.impl.LocalDateTimeDecodeHandler;
import com.serotonin.modbus4j.code.FunctionCode;
import com.serotonin.modbus4j.msg.*;
import com.serotonin.modbus4j.serial.rtu.RtuMessageParser;
import com.serotonin.modbus4j.serial.rtu.RtuMessageResponse;
import com.serotonin.modbus4j.sero.util.queue.ByteQueue;
import jakarta.annotation.Resource;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.ObjectUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.math.NumberUtils;
import org.springframework.amqp.core.BatchMessageListener;
import org.springframework.amqp.core.Message;
import org.springframework.data.redis.core.HashOperations;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.jdbc.core.JdbcTemplate;

import java.time.Duration;
import java.time.Instant;
import java.time.LocalDateTime;
import java.time.ZoneOffset;
import java.util.*;

/**
 * @author <a href="mailto:scwsl@foxmail.com">王仕龙</a>
 * 2024/11/23 0:43
 */
@Slf4j
@RequiredArgsConstructor
@SuppressWarnings("all")
public class ModbusMessagePersistListener implements BatchMessageListener {

    public static final String COMMAND_CACHE = "modbus:command:";
    private static final String COMMAND_SQL = "select * from commands where id = ?";
    private static final String COMMAND_POINT_SQL = "select * from command_points where command_id = ?";

    @Resource
    private JdbcTemplate jdbcTemplate;
    @Resource(name = "stringRedisTemplate")
    private RedisTemplate redisTemplate;

    @Resource(name = HighLowBinDecodeHandler.NAME + DecodeHandler.DECODE_NAME)
    private DecodeHandler highLowBinDecodeHandler;
    @Resource
    private Map<String, DecodeHandler> decodeHandlerMap;

    @Override
    public void onMessage(Message message) {
        this.onMessageBatch(List.of(message));
    }

    @Override
    public void onMessageBatch(List<Message> messages) {
        Long commandId;
        String[] messageSplit;
        Map<String, Object> commandMap;
        String messageString, collectionMessage, commandString;
        Map<String, ModbusMessage.MessagePoint> messagePointMap;
        // 网关标识/设备标识/命令标识/采集下发时间/采集数据接收时间/采集数据体
        // 4B454E454E4731343030303030333538/1/10000/1732365690000/1732365691000/01042C07D0000
        for (Message message : messages) {
            messageString = new String(message.getBody());
            messageSplit = StringUtils.split(messageString, "/");
            if (messageSplit.length < 6) {
                log.error("非法数据: {}", messageString);
                continue;
            }

            messagePointMap = new HashMap<>();
            collectionMessage = messageSplit[5];
            commandId = NumberUtils.createLong(messageSplit[2]);
            commandString = Objects.toString(this.redisTemplate.opsForValue().get(COMMAND_CACHE + commandId), "");
            commandMap = StringUtils.isNotEmpty(commandString) ? JSONUtil.parseObj(commandString) : null;
            if (ObjectUtils.isEmpty(commandMap)) {
                commandMap = this.jdbcTemplate.queryForMap(COMMAND_SQL, commandId);
                if (ObjectUtils.isEmpty(commandMap)) {
                    log.error("指令[{}]不存在，数据: {}", commandId, messageString);
                    continue;
                }
                this.redisTemplate.opsForValue().setIfAbsent(COMMAND_CACHE + commandId, JSONUtil.toJsonStr(commandMap), Duration.ofDays(1));
            }
            try {
                String address;
                int startAddress = MapUtil.getInt(commandMap, "start_address"), index = 0, stepSize = 0;
                ByteQueue byteQueue = new ByteQueue(collectionMessage);
                RtuMessageParser masterParser = new RtuMessageParser(true);
                RtuMessageResponse response = (RtuMessageResponse) masterParser.parseMessage(byteQueue);
                ModbusResponse modbusResponse = response.getModbusResponse();

                short[] values = null;
                switch (modbusResponse.getFunctionCode()) {
                    case FunctionCode.READ_COILS -> {
                        values = ((ReadCoilsResponse) modbusResponse).getShortData();
                    }
                    case FunctionCode.READ_DISCRETE_INPUTS -> {
                        values = ((ReadDiscreteInputsResponse) modbusResponse).getShortData();
                    }
                    case FunctionCode.READ_INPUT_REGISTERS -> {
                        values = ((ReadInputRegistersResponse) modbusResponse).getShortData();
                    }
                    case FunctionCode.READ_HOLDING_REGISTERS -> {
                        values = ((ReadHoldingRegistersResponse) modbusResponse).getShortData();
                    }
                    case FunctionCode.READ_EXCEPTION_STATUS -> {
                        values = new short[]{((ReadExceptionStatusResponse) modbusResponse).getExceptionStatus()};
                    }
                    default -> {
                        throw new RuntimeException("Funcetion code not supported: " + modbusResponse.getFunctionCode());
                    }
                }
                for (short value : values) {
                    stepSize = index * 4;
                    messagePointMap.put(StringUtils.leftPad(String.valueOf(startAddress), 4, '0'),
                            ModbusMessage.MessagePoint.builder()
                                    .parseValue(String.valueOf(value))
                                    .originalValue(StringUtils.substring(collectionMessage, 6 + stepSize, 10 + stepSize))
                                    .build());
                    index++;
                    startAddress++;
                }
            } catch (Exception e) {
                log.error("初始数据解析异常: {}", messageString, e);
                continue;
            }
            if (ObjectUtils.isNotEmpty(messagePointMap)) {
                decode(ModbusMessage.builder()
                        .commandId(commandId)
                        .message(collectionMessage)
                        .messagePointMap(messagePointMap)
                        .gatewayIdentifier(messageSplit[0])
                        .deviceId(NumberUtils.createLong(messageSplit[1]))
                        .receiveTime(NumberUtils.createLong(messageSplit[4]))
                        .collectionTime(NumberUtils.createLong(messageSplit[3]))
                        .queueName(message.getMessageProperties().getConsumerQueue())
                        .build());
            }
        }
    }


    public void decode(ModbusMessage modbusMessage) {
        Long commandId = modbusMessage.getCommandId();
        String cacheName = ModbusMessagePersistListener.COMMAND_CACHE + commandId + ":points";
        String cachePoints = Objects.toString(this.redisTemplate.opsForValue().get(cacheName), "");

        List<Map<String, Object>> commandPointList = StringUtils.isNotEmpty(cachePoints) ? (List) JSONUtil.parseArray(cachePoints) : null;
        if (ObjectUtils.isEmpty(commandPointList)) {
            commandPointList = this.jdbcTemplate.queryForList(COMMAND_POINT_SQL, commandId);
            if (ObjectUtils.isEmpty(commandPointList)) {
                log.error("指令[{}]点表配置为空", commandId, cachePoints);
                throw new RuntimeException("指令[" + commandId + "]点表未配置");
            }
            this.redisTemplate.opsForValue().setIfAbsent(cacheName, JSONUtil.toJsonStr(commandPointList), Duration.ofDays(1));
        }

        DecodeHandler decodeHandler;
        String fieldName, decodeName;
        int startAddress = 0, stepSize = 0;
        ModbusMessage.MessagePoint messagePoint;
        String deviceDataCacheName = PersistenceHandler.DEVICE_DATA_CACHE + modbusMessage.getDeviceId();
        HashOperations<String, Object, Object> hashOperations = this.redisTemplate.opsForHash();
        hashOperations.put(deviceDataCacheName, "gatewaySn", modbusMessage.getGatewayIdentifier());
        hashOperations.put(deviceDataCacheName, "deviceId", String.valueOf(modbusMessage.getDeviceId()));
        hashOperations.put(deviceDataCacheName, "receiveTime", LocalDateTime
                .ofInstant(Instant.ofEpochMilli(modbusMessage.getReceiveTime()), ZoneOffset.systemDefault())
                .format(LocalDateTimeDecodeHandler.OUT_FORMATTER));
        hashOperations.put(deviceDataCacheName, "collectionTime", LocalDateTime
                .ofInstant(Instant.ofEpochMilli(modbusMessage.getCollectionTime()), ZoneOffset.systemDefault())
                .format(LocalDateTimeDecodeHandler.OUT_FORMATTER));
        for (Map<String, Object> point : commandPointList) {
            fieldName = MapUtil.getStr(point, "field");
            stepSize = MapUtil.getInt(point, "step_size");
            decodeName = MapUtil.getStr(point, "decode_name");
            startAddress = MapUtil.getInt(point, "start_address");
            String value;
            if (StringUtils.equals(decodeName, HighLowBinDecodeHandler.NAME)) {
                value = decodeHighLowCommandPoint(modbusMessage.getMessagePointMap(), decodeName, point, startAddress);
            } else if (StringUtils.contains(decodeName, ",") && StringUtils.startsWith(decodeName, HighLowBinDecodeHandler.NAME)) {
                value = decodeHighLowStepCommandPoint(modbusMessage.getMessagePointMap(), decodeName, point, startAddress, stepSize / 2);
            } else if (stepSize <= 1) {
                messagePoint = modbusMessage.getMessagePointMap()
                        .get(StringUtils.leftPad(String.valueOf(startAddress), 4, '0'));
                decodeMessage(decodeName, point, messagePoint);
                value = messagePoint.getValue();
            } else {
                value = decodeStepCommandPoint(modbusMessage.getMessagePointMap(), decodeName, point, startAddress, stepSize);
            }
            hashOperations.put(deviceDataCacheName, fieldName, value);
        }
    }

    private void decodeMessage(String decodeName, Map<String, Object> commandPointMap, ModbusMessage.MessagePoint messagePoint) {
        if (StringUtils.isNotBlank(decodeName)) {
            DecodeHandler decodeHandler;
            if (StringUtils.contains(decodeName, ",")) {
                decodeHandler = new ComposeDecodeHandler(Arrays.stream(StringUtils.split(decodeName, ","))
                        .map(name -> this.decodeHandlerMap.get(name + DecodeHandler.DECODE_NAME)).filter(Objects::nonNull).toList());
            } else {
                decodeHandler = this.decodeHandlerMap.get(decodeName + DecodeHandler.DECODE_NAME);
            }
            if (Objects.nonNull(decodeHandler)) {
                decodeHandler.decode(commandPointMap, messagePoint);
            }
        }
    }

    private String decodeMessage(String decodeName, Map<String, Object> commandPointMap, String value) {
        if (StringUtils.isNotBlank(decodeName)) {
            DecodeHandler decodeHandler;
            if (StringUtils.contains(decodeName, ",")) {
                List<DecodeHandler> handlers = Arrays.stream(StringUtils.split(decodeName, ","))
                        .filter(name -> !StringUtils.equals(name, HighLowBinDecodeHandler.NAME))
                        .map(name -> this.decodeHandlerMap.get(name + DecodeHandler.DECODE_NAME))
                        .filter(Objects::nonNull)
                        .toList();
                if (handlers.size() == 1) {
                    decodeHandler = handlers.get(0);
                } else {
                    decodeHandler = new ComposeDecodeHandler(handlers);
                }
            } else {
                decodeHandler = this.decodeHandlerMap.get(decodeName + DecodeHandler.DECODE_NAME);
            }
            if (Objects.nonNull(decodeHandler)) {
                return decodeHandler.decode(commandPointMap, value);
            }
        }
        return value;
    }

    private String decodeStepCommandPoint(Map<String, ModbusMessage.MessagePoint> pointMap, String decodeName,
                                          Map<String, Object> commandPointMap, int startAddress, int stepSize) {
        String[] values = new String[stepSize];
        ModbusMessage.MessagePoint messagePoint;
        for (int i = 0; i < stepSize; i++) {
            messagePoint = pointMap.get(StringUtils.leftPad(String.valueOf(startAddress + i), 4, '0'));
            values[i] = messagePoint.getValue();
        }
        String format = MapUtil.getStr(commandPointMap, "format");
        return decodeMessage(decodeName, commandPointMap,
                StringUtils.isBlank(format) ? StringUtils.join(values) : String.format(format, values));
    }

    private String decodeHighLowCommandPoint(Map<String, ModbusMessage.MessagePoint> pointMap, String decodeName,
                                             Map<String, Object> commandPointMap, int startAddress) {
        String[] values = new String[2];
        ModbusMessage.MessagePoint messagePoint;
        for (int i = 0; i < 2; i++) {
            messagePoint = pointMap.get(StringUtils.leftPad(String.valueOf(startAddress + i), 4, '0'));
            values[i] = messagePoint.getOriginalValue();
        }
        return highLowBinDecodeHandler.decode(commandPointMap, StringUtils.join(values));
    }

    private String decodeHighLowStepCommandPoint(Map<String, ModbusMessage.MessagePoint> pointMap, String decodeName,
                                                 Map<String, Object> commandPointMap, int startAddress, int stepSize) {
        String[] values = new String[stepSize];
        ModbusMessage.MessagePoint highPoint;
        ModbusMessage.MessagePoint lowPoint;
        for (int i = 0; i < stepSize; i++) {
            highPoint = pointMap.get(StringUtils.leftPad(String.valueOf(startAddress + i * 2), 4, '0'));
            lowPoint = pointMap.get(StringUtils.leftPad(String.valueOf(startAddress + i * 2 + 1), 4, '0'));
            values[i] = highLowBinDecodeHandler.decode(commandPointMap, highPoint.getOriginalValue() + lowPoint.getOriginalValue());
        }

        String format = MapUtil.getStr(commandPointMap, "format");
        return decodeMessage(decodeName, commandPointMap,
                StringUtils.isBlank(format) ? StringUtils.join(values) : String.format(format, values));
    }


}
